from abc import ABC, abstractmethod
from typing import Dict, Tuple

import torch
from scipy.cluster.hierarchy import fcluster, linkage

from vidore_benchmark.utils.torch_utils import get_torch_device


class BaseEmbeddingPooler(ABC):
    """
    Abstract class for pooling embeddings.
    """

    @abstractmethod
    def pool_embeddings(self, embeddings: torch.Tensor) -> Tuple[torch.Tensor, Dict[int, torch.Tensor]]:
        """
        Return the pooled embeddings and the mapping from cluster id to token indices.
        """
        pass


class HierarchicalEmbeddingPooler(BaseEmbeddingPooler):
    """
    Hierarchical pooling of embeddings based on the similarity between tokens.
    """

    def __init__(self, pool_factor: int, device: str = "auto"):
        self.pool_factor = pool_factor
        self.device = get_torch_device(device)

    def pool_embeddings(self, embeddings: torch.Tensor) -> Tuple[torch.Tensor, Dict[int, torch.Tensor]]:
        """
        Return the pooled embeddings and the mapping from cluster id to token indices.

        Input:
        - embeddings: tensor of shape (token_length, embedding_dim)
        Output:
        - pooled_embeddings: tensor of shape (num_clusters, embedding_dim)

        NOTE: This method doesn't support batched inputs because:
        - the sequence lengths can be different.
        - scipy doesn't support batched inputs.
        """
        embeddings = embeddings.to(self.device)
        pooled_embeddings = []
        token_length = embeddings.size(0)

        if token_length == 1:
            raise ValueError("The input tensor must have more than one token.")

        similarities = torch.mm(embeddings, embeddings.t())
        if similarities.dtype == torch.bfloat16:
            similarities = similarities.to(torch.float16)
        similarities = 1 - similarities.cpu().numpy()

        Z = linkage(similarities, metric="euclidean", method="ward")  # noqa: N806
        max_clusters = max(token_length // self.pool_factor, 1)
        cluster_labels = fcluster(Z, t=max_clusters, criterion="maxclust")

        cluster_id_to_indices: Dict[int, torch.Tensor] = {}

        with torch.no_grad():
            for cluster_id in range(1, max_clusters + 1):
                cluster_indices = torch.where(torch.tensor(cluster_labels == cluster_id, device=self.device))[0]
                cluster_id_to_indices[cluster_id] = cluster_indices

                if cluster_indices.numel() > 0:
                    pooled_embedding = embeddings[cluster_indices].mean(dim=0)
                    pooled_embedding = torch.nn.functional.normalize(pooled_embedding, p=2, dim=-1)
                    pooled_embeddings.append(pooled_embedding)

            pooled_embeddings = torch.stack(pooled_embeddings, dim=0)

        return pooled_embeddings, cluster_id_to_indices



from __future__ import annotations

import math
from typing import List, Optional, cast

import torch
import torch.nn.functional as F  # noqa: N812
from PIL import Image
from torch import Tensor
from tqdm import tqdm
from transformers import AutoImageProcessor, AutoModel, AutoTokenizer

from vidore_benchmark.retrievers.utils.register_retriever import register_vision_retriever
from vidore_benchmark.retrievers.vision_retriever import VisionRetriever
from vidore_benchmark.utils.iter_utils import batched
from vidore_benchmark.utils.torch_utils import get_torch_device


@register_vision_retriever("nomic-ai/nomic-embed-vision-v1.5")
class NomicVisionRetriever(VisionRetriever):
    def __init__(self, device: str = "auto"):
        super().__init__()
        self.device = get_torch_device(device)

        self.model = AutoModel.from_pretrained("nomic-ai/nomic-embed-vision-v1.5", trust_remote_code=True).to(
            self.device
        )
        self.processor = AutoImageProcessor.from_pretrained("nomic-ai/nomic-embed-vision-v1.5")

        self.text_model = AutoModel.from_pretrained("nomic-ai/nomic-embed-text-v1.5", trust_remote_code=True).to(
            self.device
        )
        self.text_tokenizer = AutoTokenizer.from_pretrained("nomic-ai/nomic-embed-text-v1.5", trust_remote_code=True)
        self.emb_dim_query = 768
        self.emb_dim_doc = 768

    @property
    def use_visual_embedding(self) -> bool:
        return True

    @staticmethod
    def _mean_pooling(model_output: Tensor, attention_mask: Tensor) -> Tensor:
        token_embeddings = model_output[0]
        input_mask_expanded = attention_mask.unsqueeze(-1).expand(token_embeddings.size()).float()
        return torch.sum(token_embeddings * input_mask_expanded, 1) / torch.clamp(input_mask_expanded.sum(1), min=1e-9)

    def forward_queries(self, queries, batch_size: int, **kwargs) -> List[torch.Tensor]:
        list_emb_queries: List[torch.Tensor] = []
        for query_batch in tqdm(
            batched(queries, batch_size), desc="Query batch", total=math.ceil(len(queries) / batch_size)
        ):
            query_batch = cast(List[str], query_batch)

            query_texts = ["search_query: " + query for query in query_batch]
            encoded_input = self.text_tokenizer(query_texts, padding=True, truncation=True, return_tensors="pt").to(
                self.device
            )
            with torch.no_grad():
                qs = self.text_model(**encoded_input)
            qs = self._mean_pooling(qs, encoded_input["attention_mask"])  # type: ignore
            qs = F.layer_norm(qs, normalized_shape=(qs.shape[1],))
            qs = F.normalize(qs, p=2, dim=1)

            query_embeddings = torch.tensor(qs).to(self.device)
            list_emb_queries.append(query_embeddings)

        return list_emb_queries

    def forward_documents(self, documents, batch_size: int, **kwargs) -> List[torch.Tensor]:
        list_emb_documents: List[torch.Tensor] = []
        for doc_batch in tqdm(
            batched(documents, batch_size), desc="Document batch", total=math.ceil(len(documents) / batch_size)
        ):
            doc_batch = cast(List[Image.Image], doc_batch)

            vision_inputs = self.processor(doc_batch, return_tensors="pt").to(self.device)
            with torch.no_grad():
                ps = self.model(**vision_inputs).last_hidden_state
                ps = F.normalize(ps[:, 0], p=2, dim=1)

            doc_embeddings = torch.tensor(ps).to(self.device)
            list_emb_documents.append(doc_embeddings)

        return list_emb_documents

    def get_scores(
        self,
        list_emb_queries: List[torch.Tensor],
        list_emb_documents: List[torch.Tensor],
        batch_size: Optional[int] = None,
    ) -> torch.Tensor:
        emb_queries = torch.cat(list_emb_queries, dim=0)
        emb_documents = torch.cat(list_emb_documents, dim=0)

        scores = torch.einsum("bd,cd->bc", emb_queries, emb_documents)

        return scores



from typing import Any, Dict, Optional, Tuple, Union, cast

import matplotlib.pyplot as plt
import numpy as np
import numpy.typing as npt
import seaborn as sns
import torch
from PIL import Image


def plot_similarity_patches(
    img: Image.Image,
    patch_size: int,
    image_resolution: int,
    similarity_map: Optional[Union[npt.NDArray, torch.Tensor]] = None,
    figsize: Tuple[int, int] = (8, 8),
    style: Optional[Union[Dict[str, Any], str]] = None,
) -> Tuple[plt.Figure, plt.Axes]:
    """
    Plot patches of a square image.
    Set `style` to "dark_background" if your image has a light background.
    """
    max_opacity = 255

    # Get the number of patches
    if image_resolution % patch_size != 0:
        raise ValueError("The image resolution must be divisible by the patch size.")
    num_patches = image_resolution // patch_size

    # Default style
    if style is None:
        style = {}

    # Sanity checks
    if similarity_map is not None:
        if isinstance(similarity_map, torch.Tensor):
            similarity_map = cast(npt.NDArray, similarity_map.cpu().numpy())
        if similarity_map.shape != (num_patches, num_patches):
            raise ValueError("The shape of the patch_opacities tensor is not correct.")
        if not np.all((0 <= similarity_map) & (similarity_map <= 1)):
            raise ValueError("The patch_opacities tensor must have values between 0 and 1.")

    # If the image is not square, raise an error
    if img.size[0] != img.size[1]:
        raise ValueError("The image must be square.")

    # Get the image as a numpy array
    img_array = np.array(img.convert("RGBA"))  # (H, W, C) where the last channel is the alpha channel

    # Create a figure
    with plt.style.context(style):
        fig, axis = plt.subplots(num_patches, num_patches, figsize=figsize)

        # Plot the patches
        for i in range(num_patches):
            for j in range(num_patches):
                patch = img_array[i * patch_size : (i + 1) * patch_size, j * patch_size : (j + 1) * patch_size, :]
                # Set the opacity of the patch
                if similarity_map is not None:
                    patch[:, :, -1] = round(similarity_map[i, j] * max_opacity)
                axis[i, j].imshow(patch)
                axis[i, j].axis("off")

        fig.subplots_adjust(wspace=0.1, hspace=0.1)

    fig.tight_layout()

    return fig, axis


def plot_similarity_heatmap(
    img: Image.Image,
    patch_size: int,
    image_resolution: int,
    similarity_map: Union[npt.NDArray, torch.Tensor],
    figsize: Tuple[int, int] = (8, 8),
    style: Optional[Union[Dict[str, Any], str]] = None,
    show_colorbar: bool = False,
    show_axes: bool = False,
) -> Tuple[plt.Figure, plt.Axes]:
    """
    Plot a heatmap of the attention map over the image.
    The image must be square and `attention_map` must be normalized between 0 and 1.
    """

    # Get the number of patches
    if image_resolution % patch_size != 0:
        raise ValueError("The image resolution must be divisible by the patch size.")
    num_patches = image_resolution // patch_size

    # Default style
    if style is None:
        style = {}

    # Sanity checks
    if isinstance(similarity_map, torch.Tensor):
        similarity_map = cast(npt.NDArray, similarity_map.to(torch.float32).cpu().numpy())
    if similarity_map.shape != (num_patches, num_patches):
        raise ValueError("The shape of the patch_opacities tensor is not correct.")
    if not np.all((0 <= similarity_map) & (similarity_map <= 1)):
        raise ValueError("The patch_opacities tensor must have values between 0 and 1.")

    # If the image is not square, raise an error
    if img.size[0] != img.size[1]:
        raise ValueError("The image must be square.")

    # Get the image as a numpy array
    img_array = np.array(img.convert("RGBA"))  # (H, W, C) where the last channel is the alpha channel

    # Get the attention map as a numpy array
    attention_map_image = Image.fromarray((similarity_map * 255).astype("uint8")).resize(
        img.size, Image.Resampling.BICUBIC
    )

    # Create a figure
    with plt.style.context(style):
        fig, ax = plt.subplots(figsize=figsize)
        ax.imshow(img_array)
        im = ax.imshow(
            attention_map_image,
            cmap=sns.color_palette("mako", as_cmap=True),
            alpha=0.5,
        )
        if show_colorbar:
            fig.colorbar(im)
        if not show_axes:
            ax.set_axis_off()
        fig.tight_layout()

    return fig, ax



from __future__ import annotations

from typing import ClassVar, List, Optional, TypeVar, cast

import torch
from colpali_engine.models import ColPali, ColPaliProcessor
from dotenv import load_dotenv
from loguru import logger
from PIL import Image
from torch.utils.data import DataLoader, Dataset
from tqdm import tqdm

from vidore_benchmark.retrievers.utils.register_retriever import register_vision_retriever
from vidore_benchmark.retrievers.vision_retriever import VisionRetriever
from vidore_benchmark.utils.torch_utils import get_torch_device

T = TypeVar("T")
load_dotenv(override=True)


class ListDataset(Dataset[T]):
    def __init__(self, elements: List[T]):
        self.elements = elements

    def __len__(self) -> int:
        return len(self.elements)

    def __getitem__(self, idx: int) -> T:
        return self.elements[idx]


@register_vision_retriever("vidore/colpali")
class ColPaliRetriever(VisionRetriever):
    """
    ColPali Retriever that implements the model from "ColPali: Efficient Document Retrieval
    with Vision Language Models".
    """

    emb_dim_query: ClassVar[int] = 128
    emb_dim_doc: ClassVar[int] = 128

    def __init__(
        self,
        model_name: str = "vidore/colpali-v1.2",
        device: str = "auto",
    ):
        super().__init__()

        self.device = get_torch_device(device)
        logger.info(f"Using device: {self.device}")

        # Load the model and LORA adapter
        self.model = cast(
            ColPali,
            ColPali.from_pretrained(
                model_name,
                torch_dtype=torch.bfloat16,
                device_map=device,
            ),
        )

        # Load the processor
        self.processor = cast(ColPaliProcessor, ColPaliProcessor.from_pretrained(model_name))
        print("Loaded custom processor.\n")

    @property
    def use_visual_embedding(self) -> bool:
        return True

    def process_images(self, images: List[Image.Image], **kwargs):
        return self.processor.process_images(images=images)

    def process_queries(self, queries: List[str], **kwargs):
        return self.processor.process_queries(queries=queries)

    def forward_queries(self, queries: List[str], batch_size: int, **kwargs) -> List[torch.Tensor]:
        dataloader = DataLoader(
            dataset=ListDataset[str](queries),
            batch_size=batch_size,
            shuffle=False,
            collate_fn=self.process_queries,
        )

        qs = []
        for batch_query in tqdm(dataloader, desc="Forward pass queries..."):
            with torch.no_grad():
                batch_query = {k: v.to(self.device) for k, v in batch_query.items()}
                embeddings_query = self.model(**batch_query)
                qs.extend(list(torch.unbind(embeddings_query)))

        return qs

    def forward_documents(self, documents: List[Image.Image], batch_size: int, **kwargs) -> List[torch.Tensor]:
        dataloader = DataLoader(
            dataset=ListDataset[Image.Image](documents),
            batch_size=batch_size,
            shuffle=False,
            collate_fn=self.process_images,
        )

        ds = []
        for batch_doc in tqdm(dataloader, desc="Forward pass documents..."):
            with torch.no_grad():
                batch_doc = {k: v.to(self.device) for k, v in batch_doc.items()}
                embeddings_doc = self.model(**batch_doc)
            ds.extend(list(torch.unbind(embeddings_doc)))
        return ds

    def get_scores(
        self,
        list_emb_queries: List[torch.Tensor],
        list_emb_documents: List[torch.Tensor],
        batch_size: Optional[int] = 128,
    ) -> torch.Tensor:
        if batch_size is None:
            raise ValueError("`batch_size` must be provided for ColPaliRetriever's scoring")
        scores = self.processor.score(
            list_emb_queries,
            list_emb_documents,
            batch_size=batch_size,
            device=self.device,
        )
        return scores



import warnings
from typing import Optional

import numpy as np
import torch
from sentence_transformers import quantize_embeddings

from vidore_benchmark.compression.quantization import BaseEmbeddingQuantizer


class EmbeddingInt8Quantizer(BaseEmbeddingQuantizer):
    """
    Embedding quantizer that quantizes the embeddings to int8.
    """

    def __init__(
        self,
        ranges: Optional[np.ndarray] = None,
        calibration_embeddings: Optional[np.ndarray] = None,
    ):
        self.ranges = ranges
        self.calibration_embeddings = calibration_embeddings

    def set_ranges(self, ranges: np.ndarray) -> None:
        """
        Set the min/max ranges for quantization.

        Args:
        - ranges (np.ndarray): min/max ranges (2, embedding_dim).
        """
        self.ranges = ranges

    def set_calibration_embeddings(self, calibration_embeddings: np.ndarray) -> None:
        """
        Set the calibration embeddings for quantization.

        Args:
        - calibration_embeddings (np.ndarray): calibration embeddings (n_examples, embedding_dim).
        """
        self.calibration_embeddings = calibration_embeddings

    def quantize(self, embeddings: torch.Tensor) -> torch.Tensor:
        """
        Quantize the input embeddings to 8-bit integers.

        Input:
        - embeddings (torch.Tensor): input embeddings (batch_size, seq_len, embedding_dim)
        Output:
        - quantized_embeddings (torch.Tensor): quantized embeddings (batch_size, seq_len, quantized_dim)
        """

        warnings.warn(
            "EmbeddingInt8Quantizer is experimental and may change in the future.",
            UserWarning,
        )

        if self.ranges is None and self.calibration_embeddings is None:
            raise ValueError("Ranges and calibration embeddings must be set before quantizing embeddings.")
        if self.ranges is not None and self.calibration_embeddings is not None:
            raise ValueError(
                "Only one of ranges and calibration embeddings should be set before quantizing embeddings."
            )

        emb_dim = embeddings.shape[-1]
        emb_quantized = quantize_embeddings(
            embeddings.to(torch.float16).reshape(-1, emb_dim),
            precision="int8",
            ranges=self.ranges,
            calibration_embeddings=self.calibration_embeddings,
        ).reshape(embeddings.shape)

        return torch.tensor(emb_quantized, device=embeddings.device, dtype=torch.int8)



from pathlib import Path
from typing import Annotated, List, cast

import torch
import typer
from colpali_engine.models import ColPali, ColPaliProcessor
from dotenv import load_dotenv
from loguru import logger
from PIL import Image

from vidore_benchmark.interpretability.gen_similarity_maps import gen_and_save_similarity_map_per_token
from vidore_benchmark.utils.constants import OUTPUT_DIR
from vidore_benchmark.utils.logging_utils import setup_logging
from vidore_benchmark.utils.torch_utils import get_torch_device

load_dotenv(override=True)


app = typer.Typer(
    help="CLI for generating similarity maps for ColPali.",
    no_args_is_help=True,
)


@app.command()
def generate_similarity_maps(
    documents: Annotated[List[Path], typer.Option(help="List of document filepaths (image format)")],
    queries: Annotated[List[str], typer.Option(help="List of queries")],
    device: Annotated[str, typer.Option(help="Torch device")] = "auto",
    log_level: Annotated[str, typer.Option("--log", help="Logging level")] = "warning",
) -> None:
    """
    Load the ColPali model and, for each query-document pair, generate similarity maps fo
    each token in the current query.
    """

    logger.enable("vidore_benchmark")
    setup_logging(log_level)

    # Sanity checks
    assert len(documents) == len(queries), "The number of documents and queries must be the same."
    for document in documents:
        if not document.is_file():
            raise FileNotFoundError(f"File not found: `{document}`")

    # Set the device
    device = get_torch_device(device)
    print(f"Using device: {device}")

    # Load the model and LORA adapter
    model_name = "vidore/colpali-v1.2"
    processor_name = "google/paligemma-3b-mix-448"
    model = cast(
        ColPali,
        ColPali.from_pretrained(
            model_name,
            torch_dtype=torch.bfloat16,
            device_map=device,
        ),
    )

    # Load the processor
    processor = cast(ColPaliProcessor, ColPaliProcessor.from_pretrained(processor_name))
    print("Loaded custom processor.\n")

    images = [Image.open(img_filepath) for img_filepath in documents]

    for query, image, filepath in zip(queries, images, documents):
        print(f"\n\nProcessing query `{query}` and document `{filepath}`\n")
        savedir = OUTPUT_DIR / "interpretability" / filepath.stem
        gen_and_save_similarity_map_per_token(
            model=model,
            processor=processor,
            query=query,
            image=image,
            savedir=savedir,
        )

    print("\nDone.")


if __name__ == "__main__":
    app()



import glob
import os
import random
from pathlib import Path

from loguru import logger
from pdf2image import convert_from_path
from tqdm import tqdm


def convert_pdf_to_images(pdf_file: str, save_folder: str) -> None:
    """
    Convert each page of a pdf to a jpg image and save them in a folder.

    Args:
    - pdf_file (str): path to the pdf file
    - save_folder (str): path to the folder where the images will be saved

    """
    images = convert_from_path(pdf_file)

    for i, image in enumerate(images):
        if not os.path.exists(save_folder):
            os.makedirs(save_folder)
        savepath = os.path.join(save_folder, f"page_{i+1}.jpg")
        image.save(savepath, "JPEG")
        logger.debug("Saved image to `%s`", savepath)

    return


def convert_all_pdfs_to_images(path_to_folder: str, n_samples: int = 0) -> None:
    """
    Convert all PDFs in a folder and its subfolder to images and save them in a folder.
    It will sample n_samples pdf files in each subfolder, allowing to have granularity
    on the number of PDF files to convert.

    Args:
    - path_to_folder (str): path to the folder containing the pdf files
    - n_samples (int): number of pdf files to sample in each subfolder

    directory structure:
    - path_to_folder
        - subfolder1
            - pdf1
            - pdf2
            - ...
        - subfolder2
            - pdf1
            - pdf2
            - ...
        - ...
    """
    sampled_files = []

    pdf_files = glob.glob(os.path.join(path_to_folder, "*.pdf"))

    if (n_samples == 0) or (len(pdf_files) <= n_samples):
        logger.info(f"Taking all pdf files in `{path_to_folder}`")
        sampled_files.extend(pdf_files)

    else:
        logger.info(f"Taking {n_samples} pdf files in `{path_to_folder}`")
        sampled_files.extend(random.sample(pdf_files, n_samples))

    pdf_files = [str(file) for file in sampled_files]

    # Create an empty text file that will contain the file paths of the corrupted pdf files
    dirpath_corrupted = Path(path_to_folder) / "corrupted_pdf_files.txt"
    dirpath_corrupted.parent.mkdir(parents=True, exist_ok=True)

    with dirpath_corrupted.open("w") as f:
        with tqdm(total=len(pdf_files)) as pbar:
            for pdf_file in pdf_files:
                pbar.set_description(f"Processing {pdf_file}")
                relative_save_folder = os.path.join("pages_extracted", *Path(pdf_file).parts[-2:])
                if not os.path.exists(os.path.join(path_to_folder, relative_save_folder)):
                    try:
                        save_folder = os.path.join(path_to_folder, relative_save_folder)
                        convert_pdf_to_images(
                            pdf_file,
                            save_folder=save_folder,
                        )
                        logger.info("Converted `%s` to images in `%s`", pdf_file, save_folder)
                    except Exception as e:
                        logger.warning("Error converting `%s` to images: %s", pdf_file, e)
                        f.write(pdf_file)
                        f.write("\n")
                pbar.update(1)
    return



from typing import Generator

import pytest

from vidore_benchmark.retrievers.siglip_retriever import SigLIPRetriever
from vidore_benchmark.utils.testing_utils import tear_down_torch


@pytest.fixture(scope="module")
def retriever() -> Generator[SigLIPRetriever, None, None]:
    yield SigLIPRetriever()
    tear_down_torch()


@pytest.mark.slow
def test_forward_queries(retriever: SigLIPRetriever, queries_fixtures):
    embedding_queries = retriever.forward_queries(queries_fixtures, batch_size=1)
    assert len(embedding_queries) == len(queries_fixtures)


@pytest.mark.slow
def test_forward_documents(retriever: SigLIPRetriever, document_images_fixture):
    embedding_docs = retriever.forward_documents(document_images_fixture, batch_size=1)
    assert len(embedding_docs) == len(document_images_fixture)


@pytest.mark.slow
def test_get_scores(retriever: SigLIPRetriever, queries_fixtures, document_images_fixture):
    emb_query = retriever.forward_queries(queries_fixtures, batch_size=1)
    emb_doc = retriever.forward_documents(document_images_fixture, batch_size=1)
    scores = retriever.get_scores(emb_query, emb_doc)
    assert scores.shape == (len(queries_fixtures), len(document_images_fixture))



import math
from typing import List, Optional, cast

import torch
from FlagEmbedding import BGEM3FlagModel
from tqdm import tqdm

from vidore_benchmark.retrievers.utils.register_retriever import register_vision_retriever
from vidore_benchmark.retrievers.vision_retriever import VisionRetriever
from vidore_benchmark.utils.iter_utils import batched
from vidore_benchmark.utils.torch_utils import get_torch_device


@register_vision_retriever("BAAI/bge-m3")
class BGEM3Retriever(VisionRetriever):
    """
    BGEM3Retriever class to retrieve embeddings the BGE-M3 model (dense embeddings).
    """

    def __init__(self, device: str = "auto"):
        super().__init__()
        self.device = get_torch_device(device)
        self.model = BGEM3FlagModel("BAAI/bge-m3", use_fp16=True)
        self.emb_dim_query = 1024
        self.emb_dim_doc = 1024

    @property
    def use_visual_embedding(self) -> bool:
        return False

    def forward_queries(self, queries, batch_size: int, **kwargs) -> List[torch.Tensor]:
        list_emb_queries: List[torch.Tensor] = []
        for query_batch in tqdm(
            batched(queries, batch_size), desc="Query batch", total=math.ceil(len(queries) / batch_size)
        ):
            query_batch = cast(List[str], query_batch)
            with torch.no_grad():
                output = self.model.encode(query_batch, max_length=512)["dense_vecs"]
            query_embeddings = torch.tensor(output).to(self.device)
            list_emb_queries.append(query_embeddings)

        return list_emb_queries

    def forward_documents(self, documents: List[str], batch_size: int, **kwargs) -> List[torch.Tensor]:
        list_emb_documents: List[torch.Tensor] = []
        for doc_batch in tqdm(
            batched(documents, batch_size), desc="Document batch", total=math.ceil(len(documents) / batch_size)
        ):
            doc_batch = cast(List[str], doc_batch)
            with torch.no_grad():
                output = self.model.encode(doc_batch)["dense_vecs"]
            doc_embeddings = torch.tensor(output).to(self.device)
            list_emb_documents.append(doc_embeddings)
        return list_emb_documents

    def get_scores(
        self,
        list_emb_queries: List[torch.Tensor],
        list_emb_documents: List[torch.Tensor],
        batch_size: Optional[int] = None,
    ) -> torch.Tensor:
        emb_queries = torch.cat(list_emb_queries, dim=0)  # (num_queries, emb_dim_query)
        emb_documents = torch.cat(list_emb_documents, dim=0)  # (num_docs, emb_dim_doc)

        scores = torch.einsum("bd,cd->bc", emb_queries, emb_documents)  # (num_queries, num_docs)
        return scores



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Dict, List, Optional, Tuple

import torch
from datasets import Dataset
from mteb.evaluation.evaluators import RetrievalEvaluator


class VisionRetriever(ABC):
    """
    Abstract class for ViDoRe retrievers.
    """

    def __init__(self):
        pass

    @property
    @abstractmethod
    def use_visual_embedding(self) -> bool:
        """
        The child class should instantiate the `use_visual_embedding` property:
        - True if the retriever uses native visual embeddings (e.g. JINA-Clip, ColPali)
        - False if the retriever uses text embeddings and possibly VLM-generated captions (e.g. BM25).
        """
        pass

    @abstractmethod
    def forward_queries(self, queries: Any, batch_size: int, **kwargs) -> List[torch.Tensor]:
        """
        Forward pass the processed queries.

        NOTE: This method can either:
        - return a single tensor where the first dimension corresponds to the number of queries.
        - return a list of tensors where each tensor corresponds to a query.
        """
        pass

    @abstractmethod
    def forward_documents(self, documents: Any, batch_size: int, **kwargs) -> List[torch.Tensor]:
        """
        Forward pass the processed documents (i.e. page images).

        NOTE: This method can either:
        - return a single tensor where the first dimension corresponds to the number of documents.
        - return a list of tensors where each tensor corresponds to a document.
        """
        pass

    @abstractmethod
    def get_scores(
        self,
        list_emb_queries: List[torch.Tensor],
        list_emb_documents: List[torch.Tensor],
        batch_size: Optional[int] = None,
    ) -> torch.Tensor:
        """
        Get the scores between queries and documents.

        Inputs:
        - list_emb_queries: List[torch.Tensor] (n_queries, emb_dim_query)
        - list_emb_documents: List[torch.Tensor] (n_documents, emb_dim_doc)
        - batch_size: Optional[int]

        Output:
        - scores: torch.Tensor (n_queries, n_documents)
        """
        pass

    def get_relevant_docs_results(
        self,
        ds: Dataset,
        queries: List[str],
        scores: torch.Tensor,
        **kwargs,
    ) -> Tuple[Dict[str, float], Dict[str, Dict[str, float]]]:
        """
        Get the relevant documents and the results from the scores.

        NOTE: Override this method if the retriever has a different output format.

        Inputs:
        - queries: List[str]
        - documents: List[str]
        - scores: torch.Tensor (n_queries, n_documents)

        Outputs:
        - relevant_docs: Dict[str, float]
        {
            "query_0": {"doc_0": 1},
            "query_1": {"doc_1": 1},
            ...
        }
        - results: Dict[str, Dict[str, float]] with shape:
        {
            "query_0": {"doc_i": 19.125, "doc_1": 18.75, ...},
            "query_1": {"doc_j": 17.25, "doc_1": 16.75, ...},
            ...
        }
        """
        relevant_docs = {}
        results = {}

        queries2filename = {query: image_filename for query, image_filename in zip(ds["query"], ds["image_filename"])}
        passages2filename = {docidx: image_filename for docidx, image_filename in enumerate(ds["image_filename"])}

        for query, score_per_query in zip(queries, scores):
            relevant_docs[query] = {queries2filename[query]: 1}

            for docidx, score in enumerate(score_per_query):
                filename = passages2filename[docidx]
                score_passage = float(score.item())

                if query in results:
                    results[query][filename] = max(results[query].get(filename, 0), score_passage)
                else:
                    results[query] = {filename: score_passage}

        return relevant_docs, results

    def compute_metrics(
        self,
        relevant_docs: Any,
        results: Any,
        **kwargs,
    ):
        """
        Compute the MTEB metrics.

        NOTE: Override this method if the retriever has a different evaluation metric.
        """

        mteb_evaluator = RetrievalEvaluator()

        ndcg, _map, recall, precision, naucs = mteb_evaluator.evaluate(
            relevant_docs,
            results,
            mteb_evaluator.k_values,
            ignore_identical_ids=kwargs.get("ignore_identical_ids", True),
        )

        mrr = mteb_evaluator.evaluate_custom(relevant_docs, results, mteb_evaluator.k_values, "mrr")

        scores = {
            **{f"ndcg_at_{k.split('@')[1]}": v for (k, v) in ndcg.items()},
            **{f"map_at_{k.split('@')[1]}": v for (k, v) in _map.items()},
            **{f"recall_at_{k.split('@')[1]}": v for (k, v) in recall.items()},
            **{f"precision_at_{k.split('@')[1]}": v for (k, v) in precision.items()},
            **{f"mrr_at_{k.split('@')[1]}": v for (k, v) in mrr[0].items()},
            **{f"naucs_at_{k.split('@')[1]}": v for (k, v) in naucs.items()},
        }

        return scores



from typing import Dict, Type

from loguru import logger

from vidore_benchmark.retrievers.vision_retriever import VisionRetriever

VISION_RETRIEVER_REGISTRY: Dict[str, Type[VisionRetriever]] = {}


def register_vision_retriever(model_name: str):
    def decorator(cls):
        VISION_RETRIEVER_REGISTRY[model_name] = cls

        # NOTE: To see this log, use `logger.enable("vidore_benchmark")` at the very top of the script.
        logger.debug("Registered vision retriever `{}`", model_name)

        return cls

    return decorator



from __future__ import annotations

from collections import OrderedDict
from typing import Dict, List, Optional

import torch
from datasets import Dataset
from tqdm import tqdm

from vidore_benchmark.compression.quantization.embedding_binarizer import BaseEmbeddingQuantizer
from vidore_benchmark.compression.token_pooling import BaseEmbeddingPooler
from vidore_benchmark.retrievers.vision_retriever import VisionRetriever


def evaluate_dataset(
    vision_retriever: VisionRetriever,
    ds: Dataset,
    batch_query: int,
    batch_doc: int,
    batch_score: Optional[int] = None,
    embedding_quantizer: Optional[BaseEmbeddingQuantizer] = None,
    embedding_pooler: Optional[BaseEmbeddingPooler] = None,
) -> Dict[str, float]:
    """
    Evaluate the model on a given dataset using the MTEB metrics.

    NOTE: The dataset should contain the following columns:
    - query: the query text
    - image_filename: the filename of the image
    - image: the image (PIL.Image) if `use_visual_embedding` is True
    - text_description: the text description (i.e. the page caption or the text chunks) if
        `use_visual_embedding` is False
    """

    # Dataset: sanity check
    col_documents = "image" if vision_retriever.use_visual_embedding else "text_description"
    required_columns = ["query", col_documents, "image_filename"]

    if not all(col in ds.column_names for col in required_columns):
        raise ValueError(f"Dataset should contain the following columns: {required_columns}")

    # Remove `None` queries (i.e. pages for which no question was generated) and duplicates
    queries = list(set(ds["query"]))
    if None in queries:
        queries.remove(None)
        if len(queries) == 0:
            raise ValueError("All queries are None")
    documents = ds[col_documents]

    # Get the embeddings for the queries and documents
    emb_queries = vision_retriever.forward_queries(queries, batch_size=batch_query)
    emb_documents = vision_retriever.forward_documents(documents, batch_size=batch_doc)

    if embedding_quantizer is not None:
        # NOTE: Cannot use batched quantization because the sequence lengths can be different.
        # HOTFIX: CUDA doesn't support int8 operations, so we need to move the embeddings to CPU.
        emb_queries = [embedding_quantizer.quantize(emb_query.to("cpu")) for emb_query in emb_queries]
        emb_documents = [embedding_quantizer.quantize(emb_document.to("cpu")) for emb_document in emb_documents]

    if embedding_pooler is not None:
        for idx, emb_document in tqdm(enumerate(emb_documents), total=len(emb_documents), desc="Pooling embeddings..."):
            emb_document, _ = embedding_pooler.pool_embeddings(emb_document)
            emb_documents[idx] = emb_document

    # Get the similarity scores
    scores = vision_retriever.get_scores(emb_queries, emb_documents, batch_size=batch_score)

    # Get the relevant documents and results
    relevant_docs, results = vision_retriever.get_relevant_docs_results(ds, queries, scores)

    # Compute the MTEB metrics
    metrics = vision_retriever.compute_metrics(relevant_docs, results)

    return metrics


def get_top_k(
    vision_retriever: VisionRetriever,
    queries: List[str],
    emb_queries: List[torch.Tensor],
    emb_documents: List[torch.Tensor],
    file_names: List[str],
    k: int,
    batch_score: Optional[int] = None,
) -> Dict[str, OrderedDict[str, float]]:
    """
    Get the top-k documents for a given query.

    Output:
    {
        query_1: {
            document_1: score_1,
            ...
        },
        ...
    }
    """
    scores = vision_retriever.get_scores(emb_queries, emb_documents, batch_size=batch_score)
    passages2filename = {doc_idx: image_filename for doc_idx, image_filename in enumerate(file_names)}
    query_to_score = {}

    for query, score_per_query in zip(queries, scores):
        for docidx, score in enumerate(score_per_query):
            filename = passages2filename[docidx]
            score_passage = float(score.item())

            if query in query_to_score:
                query_to_score[query][filename] = max(query_to_score[query].get(filename, 0), score_passage)
            else:
                query_to_score[query] = {filename: score_passage}

    # Get the top-k documents
    for query in query_to_score:
        query_to_score[query] = OrderedDict(
            sorted(query_to_score[query].items(), key=lambda item: item[1], reverse=True)[:k]
        )

    return query_to_score



import math
from typing import List, Optional, cast

import torch
from PIL import Image
from tqdm import tqdm
from transformers import AutoModel, AutoProcessor

from vidore_benchmark.retrievers.utils.register_retriever import register_vision_retriever
from vidore_benchmark.retrievers.vision_retriever import VisionRetriever
from vidore_benchmark.utils.iter_utils import batched
from vidore_benchmark.utils.torch_utils import get_torch_device


@register_vision_retriever("google/siglip-so400m-patch14-384")
class SigLIPRetriever(VisionRetriever):
    def __init__(self, device: str = "auto"):
        super().__init__()
        self.device = get_torch_device(device)
        self.processor = AutoProcessor.from_pretrained("google/siglip-so400m-patch14-384")
        self.model = AutoModel.from_pretrained("google/siglip-so400m-patch14-384").to(self.device)
        self.model.eval()

    @property
    def use_visual_embedding(self) -> bool:
        return True

    def forward_queries(self, queries, batch_size: int, **kwargs) -> List[torch.Tensor]:
        list_emb_queries: List[torch.Tensor] = []
        for query_batch in tqdm(
            batched(queries, batch_size), desc="Query batch", total=math.ceil(len(queries) / batch_size)
        ):
            query_batch = cast(List[str], query_batch)
            inputs_queries = self.processor(
                text=query_batch, return_tensors="pt", padding="max_length", truncation=True
            ).to(self.device)
            qs = self.model.get_text_features(**inputs_queries)
            query_embeddings = torch.tensor(qs).to(self.device)
            list_emb_queries.append(query_embeddings)
        return list_emb_queries

    def forward_documents(self, documents, batch_size: int, **kwargs) -> List[torch.Tensor]:
        list_emb_documents: List[torch.Tensor] = []
        for doc_batch in tqdm(
            batched(documents, batch_size), desc="Document batch", total=math.ceil(len(documents) / batch_size)
        ):
            doc_batch = cast(List[Image.Image], doc_batch)
            list_doc = [document.convert("RGB") for document in doc_batch if isinstance(document, Image.Image)]

            input_image_processed = self.processor(images=list_doc, return_tensors="pt", padding=True).to(self.device)

            with torch.no_grad():
                ps = self.model.get_image_features(**input_image_processed)
                doc_embeddings = torch.tensor(ps).to(self.device)
                list_emb_documents.append(doc_embeddings)

        return list_emb_documents

    def get_scores(
        self,
        list_emb_queries: List[torch.Tensor],
        list_emb_documents: List[torch.Tensor],
        batch_size: Optional[int] = None,
    ) -> torch.Tensor:
        emb_queries = torch.cat(list_emb_queries, dim=0)
        emb_documents = torch.cat(list_emb_documents, dim=0)
        scores = torch.einsum("bd,cd->bc", emb_queries, emb_documents)
        return scores



import logging
import sys

from loguru import logger


def setup_logging(log_level: str = "WARNING") -> None:
    """
    Setup the logging with loguru.
    """

    numeric_level = getattr(logging, log_level.upper(), None)
    if not isinstance(numeric_level, int):
        raise ValueError(f"Invalid log level: {log_level}")

    logger.remove()  # remove the default logger
    logger.add(
        sink=sys.stderr,
        level=log_level.upper(),
        format="{time:MM/DD/YYYY HH:mm:ss} - {level} - {name}:{line} - {message}",
    )
    logger.info(f"Logging level set to {log_level.upper()}.")



from __future__ import annotations

from pathlib import Path
from typing import Any, ClassVar, Dict, List, Optional, Union

import pandas as pd


class EvalManager:
    """
    Stores evaluation results for various datasets and metrics. Supports the
    JSON output of the `VisionRetriever.compute_metric`.

    The data is stored in a pandas DataFrame with a MultiIndex for columns.
    The first level of the MultiIndex is the dataset name and the second level is the metric name.

    Usage:
    >>> eval_manager = EvalManager.from_dirpath("data/evaluation_results/")
    >>> print(eval_manager.data)
    """

    model_col: ClassVar[str] = "model"
    dataset_col: ClassVar[str] = "dataset"
    metric_col: ClassVar[str] = "metric"

    def __init__(self, data: Optional[pd.DataFrame] = None):
        if data is None:
            data = pd.DataFrame()
        self._df = data
        self._df.index = self._df.index.rename(EvalManager.model_col)

    def __str__(self) -> str:
        return self.data.__str__()

    @classmethod
    def from_dict(cls, data: Dict[Any, Any]):
        """
        Load evaluation results from a dictionary.

        Expected format:
        {
            "model1": pd.read_json(path1).T.stack(),
            "model2": pd.read_json(path2).T.stack(),
        }

        """
        df = pd.DataFrame.from_dict(data, orient="index")
        return cls(df)

    @classmethod
    def from_json(cls, path: Union[str, Path]):
        datapath = Path(path)
        if not datapath.is_file():
            raise FileNotFoundError(f"{path} is not a file")
        data = {}
        data[datapath.stem] = pd.read_json(datapath).T.stack()
        return cls.from_dict(data)

    @classmethod
    def from_multiple_json(cls, paths: Union[List[str], List[Path]]):
        data = {}
        for path in paths:
            datapath = Path(path)
            if not datapath.is_file():
                raise FileNotFoundError(f"{path} is not a file")
            data[datapath.stem] = pd.read_json(datapath).T.stack()
        return cls.from_dict(data)

    @classmethod
    def from_dir(cls, datadir: Union[str, Path]):
        datadir_ = Path(datadir)
        if not datadir_.is_dir():
            raise FileNotFoundError(f"{datadir} is not a directory")

        eval_files = list(datadir_.glob("*.json"))

        return cls.from_multiple_json(eval_files)

    @classmethod
    def from_csv(cls, path: Union[str, Path]):
        """
        Load evaluation results from a CSV file.
        """
        try:
            df = pd.read_csv(path, index_col=0, header=[0, 1])
            return cls(df)
        except Exception as e:
            print(f"Error loading {path}: {e}")
            raise e

    @property
    def data(self) -> pd.DataFrame:
        """
        Returns the evaluation results as a pandas DataFrame.
        """
        return self._df.copy()

    @property
    def models(self) -> pd.Index:
        """
        Returns the models for which there are evaluation results.
        """
        return self.data.index

    @property
    def datasets(self) -> pd.Index:
        """
        Returns the datasets for which there are evaluation results.
        """
        return self.data.columns.get_level_values(0).unique()

    @property
    def metrics(self) -> pd.Index:
        """
        Returns the metrics for which there are evaluation results.
        """
        return self.data.columns.get_level_values(1)

    @staticmethod
    def melt(df: pd.DataFrame) -> pd.DataFrame:
        """
        Melt a suitable DataFrame (e.g. returned by `get_df_for_dataset` and
        `get_df_for_metric`) into a 'long' format.
        """
        return df.T.reset_index(names=[EvalManager.dataset_col, EvalManager.metric_col]).melt(
            id_vars=[EvalManager.dataset_col, EvalManager.metric_col],
            var_name=EvalManager.model_col,
            value_name="score",
        )

    @property
    def melted(self) -> pd.DataFrame:
        """
        Returns the evaluation results as a 'melted' DataFrame.
        Useful for plotting with seaborn.
        """
        return EvalManager.melt(self.data)

    def get_df_for_model(self, model: str) -> pd.DataFrame:
        if model not in self.data.index:
            raise ValueError(f"Model {model} not found in the evaluation results")
        return self.data.loc[[model], :]  # type: ignore

    def get_df_for_dataset(self, dataset: str) -> pd.DataFrame:
        if dataset not in self.datasets:
            raise ValueError(f"Dataset {dataset} not found in the evaluation results")
        return self.data.loc[:, (dataset, slice(None))]  # type: ignore

    def get_df_for_metric(self, metric: str) -> pd.DataFrame:
        if metric not in self.metrics:
            raise ValueError(f"Metric {metric} not found in the evaluation results")
        return self.data.loc[:, (slice(None), metric)]  # type: ignore

    def sort_by_dataset(self, ascending: bool = True) -> EvalManager:
        """
        Sort the evaluation results by dataset name.
        """
        df = self.data.T.sort_index(level=0, ascending=ascending).T
        return EvalManager(df)

    def sort_by_metric(self, ascending: bool = True) -> EvalManager:
        """
        Sort the evaluation results by metric name.
        """
        df = self.data.T.sort_index(level=1, ascending=ascending).T
        return EvalManager(df)

    def sort_columns(self, ascending: bool = True) -> EvalManager:
        """
        Sort the evaluation results by dataset name and then by metric name.
        """
        df = self.data.T.sort_index(level=[0, 1], ascending=ascending).T
        return EvalManager(df)

    def to_csv(self, path: Union[str, Path]):
        """
        Save the evaluation results to a CSV file.

        Using `Evaluation.from_csv(path_to_saved_csv)` will load the evaluation results back into memory.
        """
        savepath = Path(path)
        savepath.parent.mkdir(parents=True, exist_ok=True)
        self.data.to_csv(savepath)



from typing import Any, Dict, List

from vidore_benchmark.ocr.interfaces import BoundingBox, ExtractedWord


def to_extracted_words(data: Dict[str, Any]) -> List[ExtractedWord]:
    """
    Util function to convert Tesseract output to ExtractedWord objects.
    """
    words: List[ExtractedWord] = []

    for text, x0, y0, word_w, word_h, ocr_conf in zip(
        data["text"], data["left"], data["top"], data["width"], data["height"], data["conf"]
    ):
        if text.strip():
            words.append(
                ExtractedWord(
                    text=text,
                    bbox=BoundingBox(
                        x0=x0,
                        y0=y0,
                        x1=(x0 + word_w),
                        y1=(y0 + word_h),
                    ),
                    extraction_confidence=ocr_conf,
                )
            )
    return words



try:
    from .bge_m3_retriever import BGEM3Retriever
except ImportError:
    pass
try:
    from .bge_m3_colbert_retriever import BGEM3ColbertRetriever
except ImportError:
    pass
try:
    from .bm25_retriever import BM25Retriever
except ImportError:
    pass
try:
    from .colpali_retriever import ColPaliRetriever
except ImportError:
    pass
try:
    from .dummy_retriever import DummyRetriever
except ImportError:
    pass
try:
    from .jina_clip_retriever import JinaClipRetriever
except ImportError:
    pass
try:
    from .nomic_retriever import NomicVisionRetriever
except ImportError:
    pass
try:
    from .siglip_retriever import SigLIPRetriever
except ImportError:
    pass
try:
    from .vision_retriever import VisionRetriever
except ImportError:
    pass



from dataclasses import dataclass
from typing import Dict


@dataclass
class ViTConfig:
    patch_size: int
    resolution: int

    @property
    def n_patch_per_dim(self) -> int:
        if self.resolution % self.patch_size != 0:
            raise ValueError(f"Resolution {self.resolution} is not divisible by patch size {self.patch_size}")
        return self.resolution // self.patch_size


VIT_CONFIG: Dict[str, ViTConfig] = {
    "google/siglip-so400m-patch14-384": ViTConfig(patch_size=14, resolution=384),
    "timm/ViT-SO400M-14-SigLIP-384": ViTConfig(patch_size=14, resolution=384),
    "google/paligemma-3b-mix-448": ViTConfig(
        patch_size=14, resolution=448
    ),  # based on "timm/ViT-SO400M-14-SigLIP-384" with increased resolution
    "vidore/colpaligemma-3b-pt-448-base": ViTConfig(
        patch_size=14, resolution=448
    ),  # Copied from "google/paligemma-3b-mix-448"
}



from __future__ import annotations

from dataclasses import dataclass
from typing import Tuple


@dataclass
class BoundingBox:
    """
    Bounding box data structure where (x0, y0) are the coordinates of the top-left corner.
    The x-axis is pointing to the right and the y-axis is pointing down. The pixel
    coordinates are absolute and stored as integers.
    """

    x0: int
    x1: int
    y0: int
    y1: int

    def __lt__(self, other: BoundingBox) -> bool:
        """
        Implements the lexical order (left-to-right, top-to-bottom).
        """
        if self.y0 < other.y0:
            return True
        elif self.y0 == other.y0:
            if self.x0 < other.x0:
                return True
            else:
                return False
        return False

    def __le__(self, other: BoundingBox) -> bool:
        return self < other or self == other

    @property
    def area(self) -> int:
        return (self.y1 - self.y0) * (self.x1 - self.x0)

    @property
    def height(self) -> int:
        """
        Returns the height of the box.
        """
        return self.y1 - self.y0

    @property
    def width(self) -> int:
        """
        Returns the width of the box.
        """
        return self.x1 - self.x0

    @property
    def center(self) -> Tuple[float, float]:
        """
        Returns the center coordinates of the box.
        """
        x_center = 0.5 * (self.x0 + self.x1)
        y_center = 0.5 * (self.y0 + self.y1)
        return x_center, y_center

    def is_in(self, other_box: BoundingBox) -> bool:
        return (
            self.x0 >= other_box.x0 and self.x1 <= other_box.x1 and self.y0 >= other_box.y0 and self.y1 <= other_box.y1
        )


@dataclass
class ExtractedWord:
    """
    Contains one word from the output of an OCR model (e.g. Tesseract).
    """

    text: str
    bbox: BoundingBox
    extraction_confidence: float

    def __lt__(self, other: ExtractedWord) -> bool:
        return self.bbox < other.bbox

    def __le__(self, other: ExtractedWord) -> bool:
        return self.bbox <= other.bbox



from __future__ import annotations

import math
from typing import List, Optional

import torch
from PIL import Image
from torch import Tensor

from vidore_benchmark.retrievers.utils.register_retriever import register_vision_retriever
from vidore_benchmark.retrievers.vision_retriever import VisionRetriever


@register_vision_retriever("dummy_retriever")
class DummyRetriever(VisionRetriever):
    """
    Dummy retriever for testing purposes. It generates random embeddings and scores.

    NOTE: The dummy retriever takes PIL images in its `forward_documents` method.
    """

    def __init__(
        self,
        emb_dim_query: int = 512,
        emb_dim_doc: int = 512,
    ):
        super().__init__()
        self.emb_dim_query = emb_dim_query
        self.emb_dim_doc = emb_dim_doc

    @property
    def use_visual_embedding(self) -> bool:
        return True

    def forward_queries(self, queries: List[str], batch_size: int, **kwargs) -> List[Tensor]:
        return [torch.randn(batch_size, self.emb_dim_query) for _ in range(math.ceil(len(queries) / batch_size))]

    def forward_documents(self, documents: List[Image.Image], batch_size: int, **kwargs) -> List[Tensor]:
        return [torch.randn(batch_size, self.emb_dim_doc) for _ in range(math.ceil(len(documents) / batch_size))]

    def get_scores(
        self,
        list_emb_queries: List[torch.Tensor],
        list_emb_documents: List[torch.Tensor],
        batch_size: Optional[int] = None,
    ) -> torch.Tensor:
        if batch_size is None:
            batch_size = 1
        return torch.rand(len(list_emb_queries) * batch_size, len(list_emb_documents) * batch_size)



import math
from typing import List, Optional, cast

import torch
from FlagEmbedding import BGEM3FlagModel
from tqdm import tqdm

from vidore_benchmark.evaluation.colbert_score import get_colbert_similarity
from vidore_benchmark.retrievers.utils.register_retriever import register_vision_retriever
from vidore_benchmark.retrievers.vision_retriever import VisionRetriever
from vidore_benchmark.utils.iter_utils import batched
from vidore_benchmark.utils.torch_utils import get_torch_device


@register_vision_retriever("BAAI/bge-m3-colbert")
class BGEM3ColbertRetriever(VisionRetriever):
    """
    BGEM3Retriever class to retrieve embeddings the BGE-M3 model (multi-vector embeddings + ColBERT scoring).
    """

    def __init__(self, device: str = "auto"):
        super().__init__()
        self.device = get_torch_device(device)
        self.model = BGEM3FlagModel("BAAI/bge-m3", use_fp16=True)
        self.emb_dim_query = 1024
        self.emb_dim_doc = 1024

    @property
    def use_visual_embedding(self) -> bool:
        return False

    def forward_queries(self, queries: List[str], batch_size: int, **kwargs) -> List[torch.Tensor]:
        list_emb_queries: List[torch.Tensor] = []
        for query_batch in tqdm(
            batched(queries, batch_size), desc="Query batch", total=math.ceil(len(queries) / batch_size)
        ):
            query_batch = cast(List[str], query_batch)
            with torch.no_grad():
                output = self.model.encode(
                    query_batch,
                    max_length=512,
                    return_dense=False,
                    return_sparse=False,
                    return_colbert_vecs=True,
                )["colbert_vecs"]
            list_emb_queries.extend([torch.Tensor(elt) for elt in output])
        return list_emb_queries

    def forward_documents(self, documents: List[str], batch_size: int, **kwargs) -> List[torch.Tensor]:
        list_emb_documents: List[torch.Tensor] = []
        for doc_batch in tqdm(
            batched(documents, batch_size), desc="Document batch", total=math.ceil(len(documents) / batch_size)
        ):
            doc_batch = cast(List[str], doc_batch)
            with torch.no_grad():
                output = self.model.encode(
                    doc_batch,
                    max_length=512,
                    return_dense=False,
                    return_sparse=False,
                    return_colbert_vecs=True,
                )["colbert_vecs"]
            list_emb_documents.extend([torch.Tensor(elt) for elt in output])
        return list_emb_documents

    def get_scores(
        self,
        list_emb_queries: List[torch.Tensor],
        list_emb_documents: List[torch.Tensor],
        batch_size: Optional[int] = 4,
    ) -> torch.Tensor:
        if batch_size is None:
            raise ValueError("The batch size must be specified for the ColBERT scoring.")
        scores = get_colbert_similarity(list_emb_queries, list_emb_documents, batch_size=batch_size)
        return scores



import pprint
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, Optional, Tuple, Union
from uuid import uuid4

import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns
import torch
from colpali_engine.models import ColPali, ColPaliProcessor
from einops import rearrange
from PIL import Image
from tqdm import trange

from vidore_benchmark.interpretability.plot_utils import plot_similarity_heatmap, plot_similarity_patches
from vidore_benchmark.interpretability.torch_utils import normalize_similarity_map_per_query_token
from vidore_benchmark.interpretability.vit_configs import VIT_CONFIG
from vidore_benchmark.utils.constants import OUTPUT_DIR

SUPPORTED_PLOT_KINDS = ["patches", "heatmap"]


@dataclass
class InterpretabilityInput:
    query: str
    image: Image.Image
    start_idx_token: int
    end_idx_token: int


def gen_and_save_similarity_map_per_token(
    model: ColPali,
    processor: ColPaliProcessor,
    query: str,
    image: Image.Image,
    kind: str = "heatmap",
    figsize: Tuple[int, int] = (8, 8),
    add_title: bool = True,
    style: str = "dark_background",
    savedir: Optional[Union[str, Path]] = None,
) -> None:
    """
    Generate and save the similarity maps in the `outputs` directory for each token in the query.

    NOTE: The used device is the one specified in the model.
    """

    # Sanity checks
    if model.config.name_or_path not in VIT_CONFIG:
        raise ValueError(f"`{model.config.name_or_path}` is not referenced in the VIT_CONFIG dictionary.")
    vit_config = VIT_CONFIG[model.config.name_or_path]

    # Handle savepath
    if not savedir:
        savedir = OUTPUT_DIR / "interpretability" / str(uuid4())
        print(f"No savepath provided. Results will be saved to: `{savedir}`.")
    elif isinstance(savedir, str):
        savedir = Path(savedir)
    savedir.mkdir(parents=True, exist_ok=True)

    # Resize the image to square
    input_image_square = image.resize((vit_config.resolution, vit_config.resolution))

    # Preprocess the inputs
    input_text_processed = processor.process_queries([query]).to(model.device)
    input_image_processed = processor.process_images([image]).to(model.device)

    # Forward passes
    with torch.no_grad():
        output_text = model.forward(**input_text_processed)  # (1, n_text_tokens, hidden_dim)
        output_image = model.forward(**input_image_processed)  # (1, n_patch_x * n_patch_y, hidden_dim)

    # Remove the special tokens from the output
    output_image = output_image[:, : processor.image_seq_length, :]  # (1, n_patch_x * n_patch_y, hidden_dim)

    output_image = rearrange(
        output_image, "b (h w) c -> b h w c", h=vit_config.n_patch_per_dim, w=vit_config.n_patch_per_dim
    )  # (1, n_patch_x, n_patch_y, hidden_dim)

    # Get the unnormalized attention map
    similarity_map = torch.einsum(
        "bnk,bijk->bnij", output_text, output_image
    )  # (1, n_text_tokens, n_patch_x, n_patch_y)
    similarity_map_normalized = normalize_similarity_map_per_query_token(
        similarity_map
    )  # (1, n_text_tokens, n_patch_x, n_patch_y)

    # Get text token information
    n_tokens = input_text_processed.input_ids.size(1)
    text_tokens = processor.tokenizer.tokenize(processor.decode(input_text_processed.input_ids[0]))
    print("\nText tokens:")
    pprint.pprint(text_tokens)
    print("\n")

    max_sim_scores_per_token: Dict[str, float] = {}

    for token_idx in trange(1, n_tokens - 1, desc="Iterating over tokens..."):  # exclude the <bos> and the "\n" tokens
        if kind == "patches":
            fig, axis = plot_similarity_patches(
                input_image_square,
                vit_config.patch_size,
                vit_config.resolution,
                similarity_map=similarity_map_normalized[0, token_idx, :, :],
                figsize=figsize,
                style=style,
            )
            max_sim_score = similarity_map[0, token_idx, :, :].max().item()
            max_sim_scores_per_token[f"{token_idx}: {text_tokens[token_idx]}"] = max_sim_score
            if add_title:
                fig.suptitle(
                    f"Token #{token_idx}: `{text_tokens[token_idx]}`. MaxSim score: {max_sim_score:.2f}",
                    color="white",
                    fontsize=14,
                )
        elif kind == "heatmap":
            fig, ax = plot_similarity_heatmap(
                input_image_square,
                vit_config.patch_size,
                vit_config.resolution,
                similarity_map=similarity_map_normalized[0, token_idx, :, :],
                figsize=figsize,
                style=style,
            )
            max_sim_score = similarity_map[0, token_idx, :, :].max().item()
            max_sim_scores_per_token[f"{token_idx}: {text_tokens[token_idx]}"] = max_sim_score
            if add_title:
                ax.set_title(
                    f"Token #{token_idx}: `{text_tokens[token_idx]}`. MaxSim score: {max_sim_score:.2f}",
                    fontsize=14,
                )
        else:
            raise ValueError(f"Invalid `kind` input: {kind}. Supported values are: {SUPPORTED_PLOT_KINDS}")

        savepath = savedir / f"token_{token_idx}.png"
        fig.savefig(savepath, dpi=300, bbox_inches="tight")
        print(f"Saved attention map for token {token_idx} (`{text_tokens[token_idx]}`) to `{savepath}`.\n")
        plt.close(fig)

    # Plot and save the max similarity scores per token
    with sns.axes_style("darkgrid"):
        savepath = savedir / "max_sim_scores_per_token.png"
        fig, ax = plt.subplots(figsize=(1 * len(max_sim_scores_per_token), 5))
        ser = pd.Series(max_sim_scores_per_token)
        ser.plot.bar(ax=ax)
        ax.set_xlabel("Token")
        ax.set_ylabel("Score")
        ax.set_title("Max similarity score across all patches", fontsize=14)
        fig.savefig(savepath, dpi=300, bbox_inches="tight")
        print(f"Saved max similarity scores per token to `{savepath}`.\n")
        plt.close(fig)

    return



"""
Utility functions for working with images.
"""

import base64
import io
from pathlib import Path
from typing import Union, cast

import datasets
from datasets import Dataset
from loguru import logger
from PIL import Image
from torch.utils.data import IterableDataset
from tqdm import tqdm


def scale_image(image: Image.Image, new_height: int = 1024) -> Image.Image:
    """
    Scale an image to a new height while maintaining the aspect ratio.
    """
    # Calculate the scaling factor
    width, height = image.size
    aspect_ratio = width / height
    new_width = int(new_height * aspect_ratio)

    # Resize the image
    scaled_image = image.resize((new_width, new_height))
    logger.debug("Resized image to %d * %d", new_width, new_height)

    return scaled_image


def scale_to_max_dimension(image: Image.Image, max_dimension: int = 1024) -> Image.Image:
    """
    Scale an image to a maximum dimension while maintaining the aspect ratio.
    """
    # Get the dimensions of the image
    width, height = image.size

    max_original_dimension = max(width, height)

    if max_original_dimension < max_dimension:
        return image

    # Calculate the scaling factor
    aspect_ratio = max_dimension / max_original_dimension
    new_width = int(width * aspect_ratio)
    new_height = int(height * aspect_ratio)

    # Resize the image
    scaled_image = image.resize((new_width, new_height))
    logger.debug("Resized image to %d * %d", new_width, new_height)

    return scaled_image


def get_base64_image(img: Union[str, Image.Image], add_url_prefix: bool = True) -> str:
    """
    Convert an image (from a filepath or a PIL.Image object) to a JPEG-base64 string.
    """
    if isinstance(img, str):
        img = Image.open(img)
    elif isinstance(img, Image.Image):
        pass
    else:
        raise ValueError("`img` must be a path to an image or a PIL Image object.")

    buffered = io.BytesIO()
    img.save(buffered, format="jpeg")
    b64_data = base64.b64encode(buffered.getvalue()).decode("utf-8")

    return f"data:image/jpeg;base64,{b64_data}" if add_url_prefix else b64_data


def shorten_image_path(image_path: str) -> str:
    """
    Shorten the image path to make it more readable.
    """
    full_path = Path(image_path)
    enclosing_dir = full_path.parent.parent
    return str(full_path.relative_to(enclosing_dir))


def generate_dataset_from_img_folder(path_to_folder: str) -> IterableDataset:
    """
    Generate a dataset from a folder containing JPG images.

    Args:
    - path_to_folder (str): path to the folder containing the pdf files

    Returns:
    - ds (DatasetDict): a dataset containing the questions and answers generated from the pdf files

    structure of the dataset:
    - query (str): the question generated from the image
    - image (PIL.Image): the image
    - image_filename (str): the path to the image

    """
    img_files = list(Path(path_to_folder).rglob("*.jpg")) + list(Path(path_to_folder).rglob("*.jpeg"))

    # Create a Dataset from the dictionary
    features = datasets.Features(
        {
            "image": datasets.Image(),
            "image_filename": datasets.Value("string"),
        }
    )

    def gen():
        with tqdm(total=len(img_files)) as pbar:
            for image_path in img_files:
                pbar.set_description(f"Processing {shorten_image_path(str(image_path))}")

                pil_image = Image.open(image_path)

                yield {
                    "image": pil_image,
                    "image_filename": image_path,
                }
                pbar.update(1)

    # Create the dataset from the generator
    ds = cast(IterableDataset, Dataset.from_generator(gen, features=features))

    return ds



from typing import List, Tuple

import pytest
from PIL import Image

DUMMY_TEXT = """\
Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore \
et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut \
aliquip ex ea commodo consequat. Duis aute irure dolor in reprehenderit in voluptate velit esse \
cillum dolore eu fugiat nulla pariatur.
"""


@pytest.fixture
def query_document_pairs_fixture() -> List[Tuple[str, str]]:
    return [
        (
            "What are some common outcome areas targeted by TAII for different age groups?",
            "./tests/retrievers/data/ai_sample.jpeg",
        ),
        (
            "What types of accounts or products allow investors to defer paying taxes?",
            "./tests/retrievers/data/energy_sample.jpeg",
        ),
        (
            "What is the chemical formula for the ferroelectric material Lead Zirconium Titanate (PZT)?",
            "./tests/retrievers/data/gov_sample.jpeg",
        ),
        (
            "What are some mandates for the EPA under the Pollution Prevention Act?",
            "./tests/retrievers/data/healthcare_sample.jpeg",
        ),
        (
            "Quelle partie de la production pétrolière du Kazakhstan provient de champs en mer ?",
            "./tests/retrievers/data/shift_sample.jpeg",
        ),
    ]


@pytest.fixture
def queries_fixtures(query_document_pairs_fixture) -> List[str]:
    return [query for query, _ in query_document_pairs_fixture]


@pytest.fixture
def document_filepaths_fixture(query_document_pairs_fixture) -> List[Image.Image]:
    return [doc for _, doc in query_document_pairs_fixture]


@pytest.fixture
def document_images_fixture(document_filepaths_fixture) -> List[Image.Image]:
    return [Image.open(doc) for doc in document_filepaths_fixture]


@pytest.fixture
def document_ocr_text_fixture(document_filepaths_fixture) -> List[str]:
    return [DUMMY_TEXT for _ in document_filepaths_fixture]



"""
Utility functions for interpretability.
"""

import torch


def get_torch_device(device: str = "auto") -> str:
    """
    Returns the device (string) to be used by PyTorch.

    Defaults to "auto" which will use:
    - "cuda:0" if available
    - else "mps" if available
    - "cpu" otherwise.
    """
    if device == "auto":
        if torch.cuda.is_available():
            return "cuda:0"
        elif torch.backends.mps.is_available():  # for Apple Silicon
            return "mps"
        else:
            return "cpu"
    else:
        return device



from typing import List

import torch


def get_colbert_similarity(
    emb_queries: List[torch.Tensor],
    emb_documents: List[torch.Tensor],
    batch_size: int,
) -> torch.Tensor:
    """
    Evaluate the similarity scores using the ColBERT scoring function.

    Inputs:
        - emb_queries: List of query embeddings, each of shape (n_seq, emb_dim).
        - emb_documents: List of document embeddings, each of shape (n_seq, emb_dim).
        - batch_size: Batch size for the similarity computation.
    """
    if not emb_queries or not emb_documents:
        return torch.tensor([])

    if emb_queries[0].device != emb_documents[0].device:
        raise ValueError("The device of the query and document embeddings must be the same.")

    scores = []
    for i in range(0, len(emb_queries), batch_size):
        scores_batch = []
        qs_batch = torch.nn.utils.rnn.pad_sequence(
            emb_queries[i : i + batch_size],
            batch_first=True,
            padding_value=0,
        )
        for j in range(0, len(emb_documents), batch_size):
            ps_batch = torch.nn.utils.rnn.pad_sequence(
                emb_documents[j : j + batch_size],
                batch_first=True,
                padding_value=0,
            )
            scores_batch.append(torch.einsum("bnd,csd->bcns", qs_batch, ps_batch).max(dim=3)[0].sum(dim=2))
        scores_batch = torch.cat(scores_batch, dim=1)
        scores.append(scores_batch)
    scores = torch.cat(scores, dim=0)
    return scores



import math
import warnings

import torch
from sentence_transformers import quantize_embeddings

from vidore_benchmark.compression.quantization import BaseEmbeddingQuantizer


class EmbeddingBinarizer(BaseEmbeddingQuantizer):
    """
    Embedding quantizer that binarizes the embeddings and packs them into 8-bit integers.
    """

    @staticmethod
    def pad_last_dim_to_multiple_of_8(x: torch.Tensor) -> torch.Tensor:
        """
        Pad the last dimension of the tensor to be a multiple of 8.
        """
        last_dim = x.shape[-1]
        new_last_dim = math.ceil(last_dim / 8) * 8
        padding = [0, new_last_dim - last_dim]
        padded_tensor = torch.nn.functional.pad(x, padding)
        return padded_tensor

    def quantize(self, embeddings: torch.Tensor) -> torch.Tensor:
        """
        Quantize the input embeddings using binary quantization:
        1. each element is thresholded to 0 if it is less than 0, and 1 otherwise.
        2. the thresholded embeddings are packed into 8-bit integers.

        Input:
        - embeddings (torch.Tensor): input embeddings (batch_size, seq_len, embedding_dim)
        Output:
        - quantized_embeddings (torch.Tensor): quantized embeddings (batch_size, seq_len, quantized_dim)
        """

        warnings.warn(
            "EmbeddingBinarizer is experimental and may change in the future.",
            UserWarning,
        )

        # NOTE: We pad the last dimension of the embeddings to be a multiple of 8
        # because the quantize_embeddings function takes a 2D tensor as input and pack
        # the bits into 8-bit integers (int8). To make our 3D tensors compatible, we
        # pad the last dimension to be a multiple of 8 and reshape the tensor to 2D.
        # This ensures that each int8 only contains bits from a single token.
        emb_padded = self.pad_last_dim_to_multiple_of_8(embeddings)

        batch_size, *intermediate_dims, dim = emb_padded.shape
        assert dim % 8 == 0, "The last dimension of the embeddings should be a multiple of 8."
        packed_dim = dim // 8

        emb_binarized = quantize_embeddings(
            emb_padded.to(torch.float16).reshape(batch_size, -1),
            precision="binary",
        ).reshape(batch_size, *intermediate_dims, packed_dim)

        return torch.tensor(emb_binarized, device=embeddings.device, dtype=torch.int8)



from typing import Dict, List, Optional, Union, cast

import numpy as np
import torch
from nltk.corpus import stopwords
from nltk.tokenize import word_tokenize
from PIL import Image
from rank_bm25 import BM25Okapi

from vidore_benchmark.retrievers.utils.register_retriever import register_vision_retriever
from vidore_benchmark.retrievers.vision_retriever import VisionRetriever
from vidore_benchmark.utils.torch_utils import get_torch_device


@register_vision_retriever("bm25")
class BM25Retriever(VisionRetriever):
    def __init__(self, device: str = "auto"):
        super().__init__()
        self.device = get_torch_device(device)

    @property
    def use_visual_embedding(self) -> bool:
        return False

    def forward_queries(self, queries, batch_size: int, **kwargs) -> List[torch.Tensor]:
        raise NotImplementedError("BM25Retriever only need get_scores_bm25 method.")

    def forward_documents(self, documents: List[str], batch_size: int, **kwargs) -> List[torch.Tensor]:
        raise NotImplementedError("BM25Retriever only need get_scores_bm25 method.")

    def get_scores(
        self,
        list_emb_queries: List[torch.Tensor],
        list_emb_documents: List[torch.Tensor],
        batch_size: Optional[int] = None,
        **kwargs,
    ) -> torch.Tensor:
        raise NotImplementedError("BM25Retriever only need get_scores_bm25 method.")

    def get_scores_bm25(
        self,
        queries: List[str],
        documents: Union[List[Image.Image], List[str]],
        batch_query: int,
        batch_doc: int,
        **kwargs,
    ) -> torch.Tensor:
        # Sanity check: `documents` must be a list of filepaths (strings)
        if documents and not all(isinstance(doc, str) for doc in documents):
            raise ValueError("Documents must be a list of filepaths (strings)")
        documents = cast(List[str], documents)

        queries_dict = {idx: query for idx, query in enumerate(queries)}
        tokenized_queries = self.preprocess_text(queries_dict)

        corpus = {idx: passage for idx, passage in enumerate(documents)}
        tokenized_corpus = self.preprocess_text(corpus)
        output = BM25Okapi(tokenized_corpus)

        scores = []
        for query in tokenized_queries:
            score = output.get_scores(query)
            scores.append(score)

        scores = torch.tensor(np.array(scores))  # (num_queries, num_docs)

        return scores

    def preprocess_text(self, documents: Dict[int, str]) -> List[List[str]]:
        """
        Basic preprocessing of the text data:
        - remove stopwords
        - punctuation
        - lowercase all the words.
        """
        stop_words = set(stopwords.words("english"))
        tokenized_list = [
            [word.lower() for word in word_tokenize(sentence) if word.isalnum() and word.lower() not in stop_words]
            for sentence in documents.values()
        ]
        return tokenized_list



from __future__ import annotations

import json
import time
from typing import Annotated, Dict, List, Optional, Tuple, cast

import typer
from datasets import Dataset, load_dataset
from dotenv import load_dotenv
from tqdm import tqdm

from vidore_benchmark.compression.token_pooling import HierarchicalEmbeddingPooler
from vidore_benchmark.retrievers.utils.load_retriever import load_vision_retriever_from_registry
from vidore_benchmark.retrievers.vision_retriever import VisionRetriever
from vidore_benchmark.utils.constants import OUTPUT_DIR

load_dotenv(override=True)


def evaluate_dataset_with_different_pool_factors(
    vision_retriever: VisionRetriever,
    ds: Dataset,
    pool_factors: List[int],
    batch_query: int,
    batch_doc: int,
    batch_score: Optional[int] = None,
) -> Tuple[Dict[int, Dict[str, float]], Dict[int, float]]:

    # Dataset: sanity check
    col_documents = "image" if vision_retriever.use_visual_embedding else "text_description"
    required_columns = ["query", col_documents, "image_filename"]

    if not all(col in ds.column_names for col in required_columns):
        raise ValueError(f"Dataset should contain the following columns: {required_columns}")

    # Remove `None` queries (i.e. pages for which no question was generated) and duplicates
    queries = list(set(ds["query"]))
    if None in queries:
        queries.remove(None)
        if len(queries) == 0:
            raise ValueError("All queries are None")
    documents = ds[col_documents]

    # Get the embeddings for the queries and documents
    emb_queries = vision_retriever.forward_queries(queries, batch_size=batch_query)
    emb_documents = vision_retriever.forward_documents(documents, batch_size=batch_doc)

    # Get result placeholder
    pool_factor_to_metrics: Dict[int, Dict[str, float]] = {}
    pool_factor_to_scoring_latency: Dict[int, float] = {}

    # Loop over the pool factors
    pbar = tqdm(pool_factors)
    for pool_factor in pbar:
        pbar.set_description(f"Pool factor: {pool_factor}")
        emb_documents_pooled = []

        embedding_pooler = HierarchicalEmbeddingPooler(pool_factor)
        for emb_document in tqdm(emb_documents, desc="Pooling embeddings..."):
            emb_document, _ = embedding_pooler.pool_embeddings(emb_document)
            emb_documents_pooled.append(emb_document)

        # Get the similarity scores
        start_time = time.perf_counter()
        scores = vision_retriever.get_scores(emb_queries, emb_documents_pooled, batch_size=batch_score)
        end_time = time.perf_counter()
        elapsed_time = end_time - start_time
        pool_factor_to_scoring_latency[pool_factor] = elapsed_time

        # Get the relevant documents and results
        relevant_docs, results = vision_retriever.get_relevant_docs_results(ds, queries, scores)

        # Compute the MTEB metrics
        pool_factor_to_metrics[pool_factor] = vision_retriever.compute_metrics(relevant_docs, results)

    return pool_factor_to_metrics, pool_factor_to_scoring_latency


app = typer.Typer(
    help="CLI for running experiments.",
    no_args_is_help=True,
)


@app.command()
def run_experiment(
    model_name: Annotated[str, typer.Option(help="Model name alias (tagged with `@register_vision_retriever`)")],
    pool_factors: Annotated[List[int], typer.Option(help="Pooling factors for hierarchical token pooling")],
    dataset_name: Annotated[str, typer.Option(help="HuggingFace Hub dataset name")],
    split: Annotated[str, typer.Option(help="Dataset split")] = "test",
    batch_query: Annotated[int, typer.Option(help="Batch size for query embedding inference")] = 4,
    batch_doc: Annotated[int, typer.Option(help="Batch size for document embedding inference")] = 4,
    batch_score: Annotated[Optional[int], typer.Option(help="Batch size for score computation")] = 4,
):
    # Create the vision retriever
    retriever = load_vision_retriever_from_registry(model_name)()

    # Load the dataset
    OUTPUT_DIR.mkdir(parents=True, exist_ok=True)

    dataset = cast(Dataset, load_dataset(dataset_name, split=split))
    pool_factor_to_metrics, pool_factor_to_scoring_latency = evaluate_dataset_with_different_pool_factors(
        retriever,
        dataset,
        pool_factors=pool_factors,
        batch_query=batch_query,
        batch_doc=batch_doc,
        batch_score=batch_score,
    )

    for pool_factor in pool_factors:
        metrics = {dataset_name: pool_factor_to_metrics[pool_factor]}
        savepath_metrics = (
            OUTPUT_DIR
            / f"{model_name.replace('/', '_')}"
            / f"{dataset_name.replace('/', '_')}_metrics_pool_factor_{pool_factor}.json"
        )
        savepath_metrics.parent.mkdir(parents=True, exist_ok=True)
        with open(str(savepath_metrics), "w", encoding="utf-8") as f:
            json.dump(metrics, f)
    print(f"Metrics saved in `{OUTPUT_DIR}`")

    savepath_latencies = (
        OUTPUT_DIR / f"{model_name.replace('/', '_')}" / f"{dataset_name.replace('/', '_')}_scoring_latencies.json"
    )
    savepath_latencies.parent.mkdir(parents=True, exist_ok=True)
    with open(str(savepath_latencies), "w", encoding="utf-8") as f:
        json.dump(pool_factor_to_scoring_latency, f)
    print(f"Scoring latencies saved in `{savepath_latencies}`")

    print("Done!")


if __name__ == "__main__":
    app()



from abc import ABC, abstractmethod

import torch


class BaseEmbeddingQuantizer(ABC):
    """
    Abstract class for embedding quantization.
    """

    @abstractmethod
    def quantize(self, embeddings: torch.Tensor) -> torch.Tensor:
        """
        Quantize the input embeddings.

        Input:
        - embeddings (torch.Tensor): input embeddings (batch_size, seq_len, embedding_dim)
        Output:
        - quantized_embeddings (torch.Tensor): quantized embeddings (batch_size, seq_len, quantized_dim)
        """
        pass



import math

import pytest
import torch

from vidore_benchmark.compression.quantization.embedding_binarizer import EmbeddingBinarizer


@pytest.fixture
def embedding_binarizer():
    return EmbeddingBinarizer()


def test_embedding_binarizer_initialization():
    binarizer = EmbeddingBinarizer()
    assert isinstance(binarizer, EmbeddingBinarizer)


def test_pad_last_dim_to_multiple_of_8_on_2d_vector():
    binarizer = EmbeddingBinarizer()

    # Test when padding is not needed
    x = torch.randn(2, 8)
    padded = binarizer.pad_last_dim_to_multiple_of_8(x)
    assert padded.shape == (2, 8)
    assert torch.all(padded == x)

    # Test when padding is needed
    x = torch.randn(2, 5)
    padded = binarizer.pad_last_dim_to_multiple_of_8(x)
    assert padded.shape == (2, 8)
    assert torch.all(padded[:, :5] == x)
    assert torch.all(padded[:, 5:] == 0)


def test_pad_last_dim_to_multiple_of_8_on_3d_vector():
    binarizer = EmbeddingBinarizer()

    # Test when padding is not needed
    x = torch.randn(2, 3, 8)
    padded = binarizer.pad_last_dim_to_multiple_of_8(x)
    assert padded.shape == (2, 3, 8)
    assert torch.all(padded == x)

    # Test when padding is needed
    x = torch.randn(2, 3, 5)
    padded = binarizer.pad_last_dim_to_multiple_of_8(x)
    assert padded.shape == (2, 3, 8)
    assert torch.all(padded[:, :, :5] == x)
    assert torch.all(padded[:, :, 5:] == 0)


def test_quantize_batched_embeddings(embedding_binarizer):
    # Test with embeddings that don't need padding
    embeddings = torch.randn(2, 3, 8)
    quantized = embedding_binarizer.quantize(embeddings)
    assert quantized.shape == (2, 3, 1)
    assert quantized.dtype == torch.int8

    # Test with embeddings that need padding
    embeddings = torch.randn(2, 3, 5)
    quantized = embedding_binarizer.quantize(embeddings)
    assert quantized.shape == (2, 3, 1)
    assert quantized.dtype == torch.int8


def test_quantize_values(embedding_binarizer):
    embeddings = torch.tensor(
        [
            [
                [1.0, 2.0, -1.0, -1.0],  # should get thresholded to [1, 1, 0, 0, 0, 0, 0, 0]
            ]
        ],
        dtype=torch.float32,
        device="cpu",
    )

    quantized = embedding_binarizer.quantize(embeddings)

    expected = torch.tensor([[[64]]])
    assert torch.all(quantized == expected)


@pytest.mark.parametrize(
    "batch_size,seq_length,dim",
    [
        (1, 1, 8),
        (2, 3, 16),
        (5, 10, 64),
        (3, 5, 100),
    ],
)
def test_quantize_various_shapes(embedding_binarizer, batch_size, seq_length, dim):
    embeddings = torch.randn(batch_size, seq_length, dim)
    quantized = embedding_binarizer.quantize(embeddings)
    expected_dim = math.ceil(dim / 8)
    assert quantized.shape == (batch_size, seq_length, expected_dim)


def test_quantize_device_consistency(embedding_binarizer):
    embeddings = torch.randn(2, 3, 8)
    quantized_cpu = embedding_binarizer.quantize(embeddings.cpu())
    assert quantized_cpu.device == torch.device("cpu")



from typing import Generator

import pytest

from vidore_benchmark.retrievers.bm25_retriever import BM25Retriever


@pytest.fixture(scope="module")
def retriever() -> Generator[BM25Retriever, None, None]:
    yield BM25Retriever()


@pytest.mark.slow
def test_forward_queries(retriever: BM25Retriever, queries_fixtures):
    with pytest.raises(NotImplementedError):
        _ = retriever.forward_queries(queries_fixtures, batch_size=1)


@pytest.mark.slow
def test_forward_documents(retriever: BM25Retriever, document_ocr_text_fixture):
    with pytest.raises(NotImplementedError):
        _ = retriever.forward_queries(document_ocr_text_fixture, batch_size=1)


@pytest.mark.slow
def test_get_scores(retriever: BM25Retriever, queries_fixtures, document_ocr_text_fixture):
    scores = retriever.get_scores_bm25(queries_fixtures, document_ocr_text_fixture, batch_query=1, batch_doc=1)
    assert scores.shape == (len(queries_fixtures), len(document_ocr_text_fixture))



import numpy as np
import pytest
import torch

from vidore_benchmark.compression.quantization.embedding_int8_quantizer import EmbeddingInt8Quantizer

EMBEDDING_DIM = 8
RANGES = np.array([[-1.0, 1.0]] * EMBEDDING_DIM).T


@pytest.fixture
def embedding_quantizer():
    return EmbeddingInt8Quantizer(ranges=RANGES)


def test_embedding_quantizer_initialization():
    quantizer = EmbeddingInt8Quantizer(ranges=RANGES)
    assert isinstance(quantizer, EmbeddingInt8Quantizer)


def test_quantize_batched_embeddings(embedding_quantizer):
    embeddings = torch.randn(2, 3, 8)
    quantized = embedding_quantizer.quantize(embeddings)
    assert quantized.shape == embeddings.shape
    assert quantized.dtype == torch.int8


def test_quantize_values():
    embeddings = torch.tensor(
        [
            [
                [1.0, 0.0, -1.0, -1.0],
            ]
        ],
        dtype=torch.float32,
        device="cpu",
    )

    embeddings_dim = embeddings.shape[-1]
    naive_ranges = np.array([[-1.0, 1.0]] * embeddings_dim).T
    embedding_quantizer = EmbeddingInt8Quantizer(ranges=naive_ranges)

    quantized = embedding_quantizer.quantize(embeddings)

    expected = torch.tensor([[[127, 0, -128, -128]]], dtype=torch.int8, device="cpu")
    assert torch.all(quantized == expected)


def test_quantize_device_consistency(embedding_quantizer):
    embeddings = torch.randn(2, 3, 8)
    quantized_cpu = embedding_quantizer.quantize(embeddings.cpu())
    assert quantized_cpu.device == torch.device("cpu")



from typing import Type

from vidore_benchmark.retrievers.utils.register_retriever import VISION_RETRIEVER_REGISTRY
from vidore_benchmark.retrievers.vision_retriever import VisionRetriever


def load_vision_retriever_from_registry(model_name: str) -> Type[VisionRetriever]:
    """
    Create a vision retriever instance based on the associated model name.

    To name an instance of VisionRetriever, use the following decorator:
    >>> @register_vision_retriever("my_vision_retriever")
    >>> class MyVisionRetriever(VisionRetriever):
    >>> ...
    """

    if model_name in VISION_RETRIEVER_REGISTRY:
        retriever_class = VISION_RETRIEVER_REGISTRY[model_name]
    else:
        raise ValueError(
            f"Unknown model name `{model_name}`. Available models: {list(VISION_RETRIEVER_REGISTRY.keys())}"
        )
    return retriever_class



import json
from pathlib import Path
from typing import Annotated, Optional, cast

import huggingface_hub
import numpy as np
import typer
from datasets import Dataset, load_dataset
from dotenv import load_dotenv
from loguru import logger

from vidore_benchmark.compression.quantization import EmbeddingBinarizer, EmbeddingInt8Quantizer
from vidore_benchmark.compression.token_pooling import HierarchicalEmbeddingPooler
from vidore_benchmark.evaluation.evaluate import evaluate_dataset, get_top_k
from vidore_benchmark.retrievers.colpali_retriever import ColPaliRetriever
from vidore_benchmark.retrievers.utils.load_retriever import load_vision_retriever_from_registry
from vidore_benchmark.utils.constants import OUTPUT_DIR
from vidore_benchmark.utils.image_utils import generate_dataset_from_img_folder
from vidore_benchmark.utils.logging_utils import setup_logging
from vidore_benchmark.utils.pdf_utils import convert_all_pdfs_to_images

load_dotenv(override=True)

app = typer.Typer(
    help="CLI for evaluating retrievers on the ViDoRe benchmark.",
    no_args_is_help=True,
)


@app.callback()
def main(log_level: Annotated[str, typer.Option("--log", help="Logging level")] = "warning"):
    logger.enable("vidore_benchmark")
    setup_logging(log_level)


@app.command()
def evaluate_retriever(
    model_name: Annotated[str, typer.Option(help="Model name alias (tagged with `@register_vision_retriever`)")],
    dataset_name: Annotated[Optional[str], typer.Option(help="HuggingFace Hub dataset name")] = None,
    split: Annotated[str, typer.Option(help="Dataset split")] = "test",
    batch_query: Annotated[int, typer.Option(help="Batch size for query embedding inference")] = 4,
    batch_doc: Annotated[int, typer.Option(help="Batch size for document embedding inference")] = 4,
    batch_score: Annotated[Optional[int], typer.Option(help="Batch size for score computation")] = 4,
    collection_name: Annotated[Optional[str], typer.Option(help="Collection name to use for evaluation")] = None,
    quantization: Annotated[Optional[str], typer.Option(help="Quantization method to use for embeddings")] = None,
    use_token_pooling: Annotated[bool, typer.Option(help="Whether to use token pooling for text embeddings")] = False,
    pool_factor: Annotated[int, typer.Option(help="Pooling factor for hierarchical token pooling")] = 3,
):
    """
    Evaluate the retriever on the given dataset or collection.
    The metrics are saved to a JSON file.
    """

    # Sanity check
    if dataset_name is None and collection_name is None:
        raise ValueError("Please provide a dataset name or collection name")
    elif dataset_name is not None and collection_name is not None:
        raise ValueError("Please provide only one of dataset name or collection name")

    # Create the vision retriever
    retriever = load_vision_retriever_from_registry(model_name)()

    # Get the quantization strategy
    if quantization == "binarize":
        embedding_quantizer = EmbeddingBinarizer()
    elif quantization == "int8":
        if isinstance(retriever, ColPaliRetriever):
            # ColPali's embeddings are L2-normalized so they are in the range [-1, 1]
            naive_ranges = np.array([[-1.0, 1.0]] * retriever.emb_dim_doc).T
        else:
            raise NotImplementedError("Int8 quantization is only supported for ColPali retriever.")
        embedding_quantizer = EmbeddingInt8Quantizer(ranges=naive_ranges)
    elif quantization is None:
        embedding_quantizer = None
    else:
        raise ValueError(f"Invalid quantization method: {quantization}")

    # Get the pooling strategy
    embedding_pooler = HierarchicalEmbeddingPooler(pool_factor) if use_token_pooling else None

    # Create the output directory if it doesn't exist
    OUTPUT_DIR.mkdir(parents=True, exist_ok=True)

    # Load the dataset
    if dataset_name is not None:
        dataset = cast(Dataset, load_dataset(dataset_name, split=split))
        metrics = {
            dataset_name: evaluate_dataset(
                retriever,
                dataset,
                batch_query=batch_query,
                batch_doc=batch_doc,
                batch_score=batch_score,
                embedding_quantizer=embedding_quantizer,
                embedding_pooler=embedding_pooler,
            )
        }

        if use_token_pooling:
            savepath = OUTPUT_DIR / f"{model_name.replace('/', '_')}_metrics_pool_factor_{pool_factor}.json"
        else:
            savepath = OUTPUT_DIR / f"{model_name.replace('/', '_')}_metrics.json"

        with open(str(savepath), "w", encoding="utf-8") as f:
            json.dump(metrics, f)

        print(f"Metrics saved to `{savepath}`")
        print(f"NDCG@5 for {model_name} on {dataset_name}: {metrics[dataset_name]['ndcg_at_5']}")

    elif collection_name is not None:
        collection = huggingface_hub.get_collection(collection_name)
        datasets = collection.items

        metrics_all = {}
        savedir = OUTPUT_DIR / model_name.replace("/", "_")
        savedir.mkdir(parents=True, exist_ok=True)

        for dataset_item in datasets:
            print(f"\n---------------------------\nEvaluating {dataset_item.item_id}")
            dataset = cast(Dataset, load_dataset(dataset_item.item_id, split=split))
            metrics = {
                dataset_item.item_id: evaluate_dataset(
                    retriever,
                    dataset,
                    batch_query=batch_query,
                    batch_doc=batch_doc,
                    batch_score=batch_score,
                    embedding_quantizer=embedding_quantizer,
                    embedding_pooler=embedding_pooler,
                )
            }
            metrics_all.update(metrics)

            if use_token_pooling:
                savepath = savedir / f"{dataset_item.item_id.replace('/', '_')}_metrics_pool_factor_{pool_factor}.json"
            else:
                savepath = savedir / f"{dataset_item.item_id.replace('/', '_')}_metrics.json"

            with open(str(savepath), "w", encoding="utf-8") as f:
                json.dump(metrics, f)

            print(f"Metrics saved to `{savepath}`")
            print(f"NDCG@5 for {model_name} on {dataset_item.item_id}: {metrics[dataset_item.item_id]['ndcg_at_5']}")

        if use_token_pooling:
            savepath_all = OUTPUT_DIR / f"{model_name.replace('/', '_')}_all_metrics_pool_factor_{pool_factor}.json"
        else:
            savepath_all = OUTPUT_DIR / f"{model_name.replace('/', '_')}_all_metrics.json"

        with open(str(savepath_all), "w", encoding="utf-8") as f:
            json.dump(metrics_all, f)

        print(f"Concatenated metrics saved to `{savepath_all}`")

    else:
        raise ValueError("Please provide a dataset name or collection name.")

    print("Done.")


@app.command()
def retrieve_on_dataset(
    model_name: Annotated[str, typer.Option(help="Model name alias (tagged with `@register_vision_retriever`)")],
    query: Annotated[str, typer.Option(help="Query to use for retrieval")],
    k: Annotated[int, typer.Option(help="Number of documents to retrieve")],
    dataset_name: Annotated[str, typer.Option(help="HuggingFace Hub dataset name")],
    split: Annotated[str, typer.Option(help="Dataset split")] = "test",
    batch_doc: Annotated[int, typer.Option(help="Batch size for document embedding inference")] = 4,
    batch_score: Annotated[Optional[int], typer.Option(help="Batch size for score computation")] = 4,
):
    """
    Retrieve the top-k documents according to the given query.
    """

    # Create the vision retriever
    retriever = load_vision_retriever_from_registry(model_name)()

    # Load the dataset
    ds = cast(Dataset, load_dataset(dataset_name, split=split))

    # Get embeddings for the queries and documents
    emb_queries = retriever.forward_queries([query], batch_size=1)
    emb_documents = retriever.forward_documents(
        list(ds["image"]) if retriever.use_visual_embedding else list(ds["text_description"]),
        batch_size=batch_doc,
    )

    # Get the top-k documents
    top_k = get_top_k(
        retriever,
        queries=[query],
        emb_queries=emb_queries,
        emb_documents=emb_documents,
        file_names=list(ds["image_filename"]),
        k=k,
        batch_score=batch_score,
    )

    print(f"Top-{k} documents for the query '{query}':")
    for document, score in top_k[query].items():
        print(f"- Document `{document}` (score = {score})")

    print("Done.")


@app.command()
def retrieve_on_pdfs(
    model_name: Annotated[str, typer.Option(help="Model name to use for evaluation")],
    query: Annotated[str, typer.Option(help="Query to use for retrieval")],
    k: Annotated[int, typer.Option(help="Number of documents to retrieve")],
    data_dirpath: Annotated[
        str, typer.Option(help="Path to the folder containing the PDFs to use as the retrieval corpus")
    ],
    batch_doc: Annotated[int, typer.Option(help="Batch size for document embedding inference")] = 4,
    batch_score: Annotated[Optional[int], typer.Option(help="Batch size for score computation")] = 4,
):
    """
    This script is used to ask a query and retrieve the top-k documents from a given folder containing PDFs.
    The PDFs will be converted to a dataset of image pages and then used for retrieval.
    """

    if not Path(data_dirpath).is_dir():
        raise FileNotFoundError(f"Invalid data directory: `{data_dirpath}`")

    # Create the vision retriever
    retriever = load_vision_retriever_from_registry(model_name)()

    # Convert the PDFs to a collection of images
    convert_all_pdfs_to_images(data_dirpath)
    image_files = list(Path(data_dirpath).rglob("*.jpg"))
    print(f"Found {len(image_files)} images in the directory `{data_dirpath}`")

    # Generate a dataset using the images
    ds = generate_dataset_from_img_folder(data_dirpath)

    # Get embeddings for the queries and documents
    emb_queries = retriever.forward_queries([query], batch_size=1)
    emb_documents = retriever.forward_documents(
        list(ds["image"]),
        batch_size=batch_doc,
    )

    # Get the top-k documents
    top_k = get_top_k(
        retriever,
        queries=[query],
        emb_queries=emb_queries,
        emb_documents=emb_documents,
        file_names=list(ds["image_filename"]),
        k=k,
        batch_score=batch_score,
    )

    print(f"Top-{k} documents for the query '{query}':")

    for document, score in top_k[query].items():  # type: ignore
        print(f"Document: {document}, Score: {score}")

    print("Done.")


if __name__ == "__main__":
    app()



import sys


def islice(iterable, *args):
    """
    Yield a slice of an iterable.
    >>> islice('ABCDEFG', 2) → A B
    >>> islice('ABCDEFG', 2, 4) → C D
    >>> islice('ABCDEFG', 2, None) → C D E F G
    >>> islice('ABCDEFG', 0, None, 2) → A C E G
    """
    s = slice(*args)
    start, stop, step = s.start or 0, s.stop or sys.maxsize, s.step or 1
    it = iter(range(start, stop, step))
    try:
        nexti = next(it)
    except StopIteration:
        # Consume *iterable* up to the *start* position.
        for i, element in zip(range(start), iterable):
            pass
        return
    try:
        for i, element in enumerate(iterable):
            if i == nexti:
                yield element
                nexti = next(it)
    except StopIteration:
        # Consume to *stop*.
        for i, element in zip(range(i + 1, stop), iterable):
            pass


def batched(iterable, n: int):
    """
    Yield batches of n elements from an iterable.
    >>> batched('ABCDEFG', 3) → ABC DEF G
    """
    if n < 1:
        raise ValueError("n must be at least one")
    it = iter(iterable)
    while batch := tuple(islice(it, n)):
        yield batch



import pytest
import torch

from vidore_benchmark.compression.token_pooling import HierarchicalEmbeddingPooler


@pytest.fixture
def sample_embeddings() -> torch.Tensor:
    return torch.tensor(
        [
            [1.0, 0.0, 0.0],
            [0.0, 1.0, 0.0],
            [0.0, 0.0, 1.0],
            [1.0, 0.0, 0.0],
            [1.0, 0.0, 0.0],
            [1.0, 0.0, 0.0],
        ],
        dtype=torch.float32,
    )


def test_hierarchical_embedding_pooler_initialization():
    pooler = HierarchicalEmbeddingPooler(pool_factor=2, device="cpu")
    assert pooler.pool_factor == 2
    assert pooler.device == "cpu"


def test_hierarchical_embedding_pooler_output_shape(sample_embeddings: torch.Tensor):
    pooler = HierarchicalEmbeddingPooler(pool_factor=2, device="cpu")
    pooled_embeddings, cluster_id_to_indices = pooler.pool_embeddings(sample_embeddings)

    assert isinstance(pooled_embeddings, torch.Tensor)
    assert pooled_embeddings.shape[1] == sample_embeddings.shape[1]
    assert pooled_embeddings.shape[0] <= len(cluster_id_to_indices)


def test_hierarchical_embedding_pooler_output_values(sample_embeddings: torch.Tensor):
    pooler = HierarchicalEmbeddingPooler(pool_factor=2, device="cpu")
    pooled_embeddings, cluster_id_to_indices = pooler.pool_embeddings(sample_embeddings)

    expected_pooled_embeddings = torch.tensor(
        [
            [1.0, 0.0, 0.0],
            [0.0, 1.0, 0.0],
            [0.0, 0.0, 1.0],
        ],
        dtype=torch.float32,
    )
    expected_cluster_id_to_indices = {
        1: torch.tensor([0, 3, 4, 5]),
        2: torch.tensor([1]),
        3: torch.tensor([2]),
    }

    assert torch.allclose(pooled_embeddings, expected_pooled_embeddings)
    assert all(
        [
            torch.allclose(cluster_id_to_indices[cluster_id], expected_cluster_indices)
            for cluster_id, expected_cluster_indices in expected_cluster_id_to_indices.items()
        ]
    )


def test_hierarchical_embedding_pooler_with_different_pool_factors(sample_embeddings: torch.Tensor):
    for pool_factor in [1, 2, 3]:
        pooler = HierarchicalEmbeddingPooler(pool_factor=pool_factor, device="cpu")
        pooled_embeddings, cluster_id_to_indices = pooler.pool_embeddings(sample_embeddings)
        expected_num_clusters = (sample_embeddings.shape[0] + pool_factor - 1) // pool_factor
        assert pooled_embeddings.shape[0] <= expected_num_clusters
        assert pooled_embeddings.shape[0] <= len(cluster_id_to_indices)


def test_hierarchical_embedding_pooler_should_raise_error_with_single_token():
    single_token_embeddings = torch.rand(1, 768)
    pooler = HierarchicalEmbeddingPooler(pool_factor=2, device="cpu")

    with pytest.raises(ValueError):
        pooler.pool_embeddings(single_token_embeddings)


def test_hierarchical_embedding_pooler_large_input():
    large_embeddings = torch.rand(1000, 768)

    pooler = HierarchicalEmbeddingPooler(pool_factor=10, device="cpu")
    pooled_embeddings, cluster_id_to_indices = pooler.pool_embeddings(large_embeddings)

    assert pooled_embeddings.shape[0] < large_embeddings.shape[0]
    assert pooled_embeddings.shape[0] <= len(cluster_id_to_indices)



import torch

EPSILON = 1e-10


def normalize_similarity_map_per_query_token(x: torch.Tensor) -> torch.Tensor:
    """
    Normalizes the attention map for ColPali for each query token.
    The output tensor will have values in the range [0, 1] and the
    same shape as the input tensor.

    Args:
        x: The attention map tensor of shape (batch_size, n_text_tokens, n_patch_x, n_patch_y).
    """
    if x.ndim != 4:
        raise ValueError("The input tensor must have 4 dimensions.")

    # Compute the minimum values along the last two dimensions (n_patch_x, n_patch_y)
    min_vals = x.min(dim=-1, keepdim=True)[0].min(dim=-2, keepdim=True)[0]

    # Compute the maximum values along the last two dimensions (n_patch_x, n_patch_y)
    max_vals = x.max(dim=-1, keepdim=True)[0].max(dim=-2, keepdim=True)[0]

    # Normalize the tensor
    x_normalized = (x - min_vals) / (max_vals - min_vals + EPSILON)  # Adding a small epsilon to avoid division by zero

    return x_normalized
